/*

  $Id: genall.c,v 1.11 2008-03-12 10:09:01 schuppan Exp $

  Name:    genall.c
  Version: 0.11
  Date:    28.10.2002

  Author:  Viktor Schuppan, ETH Zuerich

  Description:

  Generates modules for a model of the FireWire TIP protocol with a
  specific number of nodes, ports and determinism.

  Command-line

  genall [-n n] [-p p] [-frn] [-cfn]

  -n gives the number of nodes,
  -p gives the number of ports,
  -frn enables force-root non-determinism, 
  -cfn enables configuration non-determinism

 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#define VERSION "genall v0.12 29.08.2002"

#define MAXN 100

#define FRN 1
#define CFN 2

#define OBJDIR "obj"
#define INCDIR "lib"

void displayUsage(char *progname) {

  fprintf(stderr, "usage: %s [-n n] [-p p] [-frn] [-cfn]\n", progname); 

} /* displayUsage */


void generatePeer(int n, int p, int d) {

  FILE *file;
  char filename[256], pathname[256], symbol[256];
  int i;

  sprintf(filename, "peer_%u_%u_%x.smv", n, p, d);
  sprintf(pathname, "%s/%s", OBJDIR, filename);
  strcpy(symbol, filename);

  file = fopen(pathname, "w");
  if (file == NULL) {
    fprintf(stderr, "cannot open file %s for writing\n", filename);
    exit(1);
  }
  else
    fprintf(stderr, "generating file %s\n", filename);

  fprintf(stderr, "generating header...");
  for (i = 0; i < strlen(symbol); i++) {
    if (islower(symbol[i]))
      symbol[i] = toupper(symbol[i]);
    else if (symbol[i] == '.')
      symbol[i] = '_';
  }
  fprintf(file, "#ifndef %s\n", symbol);
  fprintf(file, "#define %s\n", symbol);
  fprintf(file, "\n");
  fprintf(file, "--\n");
  fprintf(file, "-- %s\n", filename);
  fprintf(file, "--\n");
  fprintf(file, "\n");
  fprintf(stderr, " finished\n");

  fprintf(stderr, "generating body...");
  fprintf(file, "MODULE peer\n");
  fprintf(file, "\n");
  fprintf(file, "VAR\n");
  fprintf(file, "  node: -1..%u;\n", n - 1);
  fprintf(file, "  port: 0..%u;\n", p - 1);
  fprintf(file, "ASSIGN\n");
  fprintf(file, "  next(node) := node;\n");
  fprintf(file, "  next(port) := port;\n");
  fprintf(file, "\n");
  fprintf(file, "#endif\n");
  fprintf(stderr, " finished\n");

  fclose(file);

}


int getForceRootCycles(int n, int p, int d) {

  return 3 * (n - 2) + 1;

} /* getForceRootCycles */


int getTimeoutCycles(int n, int p, int d) {

  return getForceRootCycles(n, p, d) + 3 * (n - 1) / 2 + 1;

} /* getTimeoutCycles */


int getRangeCycles(int n, int p, int d) {

  return getTimeoutCycles(n, p, d) + 1;

} /* getRangeCycles */


void printVars(FILE *file, int n, int p, int d) {

  fprintf(file, "VAR\n");
  fprintf(file, "  port: array 0..%u of port;\n", p - 1);
  fprintf(file, "  state: {T0_start, T1_child, T2_parent, T3_root_1, T3_root_2, T3_root_3, S0_start};\n");
  fprintf(file, "  root: boolean;\n");
  fprintf(file, "  parent_port: 0..%u;\n", p - 1);
  fprintf(file, "  force_root: boolean;\n");
  fprintf(file, "  rnd: boolean;\n");
  fprintf(file, "  timeout: boolean;\n");
  fprintf(file, "  cycles: 0..%u;\n", getRangeCycles(n, p, d));
  fprintf(file, "\n");
  fprintf(file, "DEFINE\n");
  fprintf(file, "  const_timeout_cycles := %u;\n", getTimeoutCycles(n, p, d));
  fprintf(file, "  const_force_root_cycles := %u;\n", getForceRootCycles(n, p, d));
  fprintf(file, "\n");

} /* printVars */


void printTrans(FILE *file, int n, int p, int d) {

  int j, j2;

  fprintf(file, "ASSIGN\n");
  fprintf(file, "  init(state) := T0_start;\n");
  fprintf(file, "  init(timeout) := 0;\n");
  fprintf(file, "  init(cycles) := 0;\n");
  fprintf(file, "\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "  -- state\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "\n");
  fprintf(file, "  next(state) := case\n");
  fprintf(file, "\n");
  fprintf(file, "    ----------------------------------------------------------------------\n");
  fprintf(file, "    -- T0_start\n");
  fprintf(file, "    ----------------------------------------------------------------------\n");
  fprintf(file, "\n");
  fprintf(file, "    -- at least n - 1 ports not unknown\n");
  fprintf(file, "    state = T0_start & (\n");
  for (j = 0; j < p; j++) {
    fprintf(file, "      (");
    for (j2 = 0; j2 < p; j2++) {
      if (j2 != j) {
	fprintf(file, "!(port[%u].role = unknown) %c ", j2, (j2 < p - 2) || ((j2 == p - 2) && (j != p - 1)) ? '&' : ' ');
      }
    }
    fprintf(file, ") %c\n", (j < p - 1) ? '|' : ' ');
  }
  fprintf(file, "    ) & \n");
  fprintf(file, "    (force_root -> cycles > const_force_root_cycles): T1_child;\n");
  fprintf(file, "\n");
  fprintf(file, "    ----------------------------------------------------------------------\n");
  fprintf(file, "    -- T1_child\n");
  fprintf(file, "    ----------------------------------------------------------------------\n");
  fprintf(file, "\n");
  fprintf(file, "    -- each child port cin is idle\n");
  fprintf(file, "    state = T1_child & \n");
  for (j = 0; j < p; j++) {
    fprintf(file, "      (port[%u].role = child -> port[%u].cin = idle) %c\n", j, j, (j != p - 1) ? '&' : ' ');
  }
  fprintf(file, "    : T2_parent;\n");
  fprintf(file, "\n");
  fprintf(file, "    ----------------------------------------------------------------------\n");
  fprintf(file, "    -- T2_parent\n");
  fprintf(file, "    ----------------------------------------------------------------------\n");
  fprintf(file, "\n");
  fprintf(file, "    state = T2_parent &\n");
  for (j = 0; j < p; j++) {
    fprintf(file, "      (port[%u].role = unknown -> port[%u].cin = child_notify) %c\n", j, j, (j != p - 1) ? '&' : ' ');
  }
  fprintf(file, "    : S0_start;\n");
  fprintf(file, "    state = T2_parent &\n");
  for (j = 0; j < p; j++) {
    fprintf(file, "      (port[%u].role = unknown -> port[%u].cin = parent_notify) %c\n", j, j, (j != p - 1) ? '&' : ' ');
  }
  fprintf(file, "    : T3_root_1;\n");
  fprintf(file, "\n");
  fprintf(file, "    ----------------------------------------------------------------------\n");
  fprintf(file, "    -- T3_root_1\n");
  fprintf(file, "    ----------------------------------------------------------------------\n");
  fprintf(file, "\n");
  fprintf(file, "    state = T3_root_1 & !rnd: T3_root_3;\n");
  fprintf(file, "    state = T3_root_1 & rnd: T3_root_2;\n");
  fprintf(file, "\n");
  fprintf(file, "    ----------------------------------------------------------------------\n");
  fprintf(file, "    -- T3_root_2\n");
  fprintf(file, "    ----------------------------------------------------------------------\n");
  fprintf(file, "\n");
  fprintf(file, "    state = T3_root_2: T3_root_3;\n");
  fprintf(file, "\n");
  fprintf(file, "    ----------------------------------------------------------------------\n");
  fprintf(file, "    -- T3_root_3\n");
  fprintf(file, "    ----------------------------------------------------------------------\n");
  fprintf(file, "\n");
  fprintf(file, "    state = T3_root_3 &\n");
  for (j = 0; j < p; j++) {
    fprintf(file, "      (port[%u].role = unknown -> port[%u].cin = idle) %c\n", j, j, (j != p - 1) ? '&' : ' ');
  }
  fprintf(file, "    : T2_parent;\n");
  fprintf(file, "    state = T3_root_3 &\n");
  for (j = 0; j < p; j++) {
    fprintf(file, "      (port[%u].role = unknown -> port[%u].cin = parent_notify) %c\n", j, j, (j != p - 1) ? '&' : ' ');
  }
  fprintf(file, "    : T1_child;\n");
  fprintf(file, "\n");
  fprintf(file, "    ----------------------------------------------------------------------\n");
  fprintf(file, "    -- else\n");
  fprintf(file, "    ----------------------------------------------------------------------\n");
  fprintf(file, "\n");
  fprintf(file, "    1: state;\n");
  fprintf(file, "\n");
  fprintf(file, "  esac; --state\n");
  fprintf(file, "\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "  -- root\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "\n");
  fprintf(file, "  next(root) := case\n");
  fprintf(file, "\n");
  fprintf(file, "    state = T1_child & (\n");
  for (j = 0; j < p; j++) {
    fprintf(file, "      port[%u].role = unknown %c\n", j, (j != p - 1) ? '|' : ' ');
  }
  fprintf(file, "    ): 0;\n");
  fprintf(file, "    state = T1_child: 1;\n");
  fprintf(file, "    1: root;\n");
  fprintf(file, "\n");
  fprintf(file, "  esac; --root\n");
  fprintf(file, "\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "  -- parent_port\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "\n");
  fprintf(file, "  next(parent_port) := case\n");
  fprintf(file, "\n");
  for (j = 0; j < p; j++) {
    fprintf(file, "    state = T1_child & port[%u].role = unknown: %u;\n", j, j);
  }
  fprintf(file, "    1: parent_port;\n");
  fprintf(file, "\n");
  fprintf(file, "  esac; --parent_port\n");
  fprintf(file, "\n");
  for (j = 0; j < p; j++) {
    fprintf(file, "  ------------------------------------------------------------------------\n");
    fprintf(file, "  ------------------------------------------------------------------------\n");
    fprintf(file, "  -- port[%u].cout\n", j);
    fprintf(file, "  ------------------------------------------------------------------------\n");
    fprintf(file, "  ------------------------------------------------------------------------\n");
    fprintf(file, "\n");
    fprintf(file, "  next(port[%u].cout) := case\n", j);
    fprintf(file, "\n");
    fprintf(file, "    -- T1_child\n");
    fprintf(file, "    state = T1_child & port[%u].role = child  : child_notify;\n", j);
    fprintf(file, "    state = T1_child & port[%u].role = unknown: parent_notify;\n", j);
    fprintf(file, "\n");
    fprintf(file, "    -- T3_root_1\n");
    fprintf(file, "    state = T3_root_1 & port[%u].role = unknown: idle;\n", j);
    fprintf(file, "\n");
    fprintf(file, "    -- T3_root_3\n");
    fprintf(file, "    state = T3_root_3 & port[%u].role = unknown & port[%u].cin = idle: parent_notify;\n", j, j);
    fprintf(file, "\n");
    fprintf(file, "    -- S0_start\n");
    fprintf(file, "    state = S0_start & port[%u].role = parent & port[%u].cin = child_notify: idle;\n", j, j);
    fprintf(file, "    state = S0_start & port[%u].role = child & port[%u].cin =idle: idle;\n", j, j);
    fprintf(file, "    state = S0_start & root: idle;\n");
    fprintf(file, "\n");
    fprintf(file, "    -- else\n");
    fprintf(file, "    1: port[%u].cout;\n", j);
    fprintf(file, "\n");
    fprintf(file, "  esac; --port[%u].cout\n", j);
    fprintf(file, "\n");
    fprintf(file, "  ------------------------------------------------------------------------\n");
    fprintf(file, "  ------------------------------------------------------------------------\n");
    fprintf(file, "  -- port[%u].role\n", j);
    fprintf(file, "  ------------------------------------------------------------------------\n");
    fprintf(file, "  ------------------------------------------------------------------------\n");
    fprintf(file, "\n");
    fprintf(file, "  next(port[%u].role) := case\n", j);
    fprintf(file, "\n");
    fprintf(file, "    state = T0_start & port[%u].role = unknown & port[%u].cin = parent_notify: child;\n", j, j);
    fprintf(file, "    state = T2_parent & port[%u].role = unknown & port[%u].cin = child_notify: parent;\n", j, j);
    fprintf(file, "    state = T3_root_3 & port[%u].role = unknown & port[%u].cin = parent_notify: child;\n", j, j);
    fprintf(file, "    1: port[%u].role;\n", j);
    fprintf(file, "\n");
    fprintf(file, "  esac; --port[%u].role\n", j);
    fprintf(file, "\n");
  }
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "  -- force_root\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "\n");
  fprintf(file, "  next(force_root) := force_root;\n");
  fprintf(file, "\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "  -- rnd\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "\n");
  fprintf(file, "  next(rnd) := case\n");
  fprintf(file, "    state = T2_parent: {0, 1};\n");
  fprintf(file, "    1: rnd;\n");
  fprintf(file, "  esac; --rnd\n");
  fprintf(file, "\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "  -- timeout\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "\n");
  fprintf(file, "  next(timeout) := timeout | state = T0_start & cycles > const_timeout_cycles;\n");
  fprintf(file, "\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "  -- cycles\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "  ------------------------------------------------------------------------\n");
  fprintf(file, "\n");
  fprintf(file, "  next(cycles) := case\n");
  fprintf(file, "\n");
  fprintf(file, "    state = T0_start: (cycles + 1) mod %u;\n", getRangeCycles(n, p, d) + 1);
  fprintf(file, "    1: cycles;\n");
  fprintf(file, "\n");
  fprintf(file, "  esac; --cycles\n");

} /* printTrans */


void generateNode(int n, int p, int d) {

  FILE *file;
  char filename[256], pathname[256], symbol[256];
  int i;

  sprintf(filename, "node_%u_%u_%x.smv", n, p, d);
  sprintf(pathname, "%s/%s", OBJDIR, filename);
  strcpy(symbol, filename);

  file = fopen(pathname, "w");
  if (file == NULL) {
    fprintf(stderr, "cannot open file %s for writing\n", filename);
    exit(1);
  }
  else
    fprintf(stderr, "generating file %s\n", filename);

  fprintf(stderr, "generating header...");
  for (i = 0; i < strlen(symbol); i++) {
    if (islower(symbol[i]))
      symbol[i] = toupper(symbol[i]);
    else if (symbol[i] == '.')
      symbol[i] = '_';
  }
  fprintf(file, "#ifndef %s\n", symbol);
  fprintf(file, "#define %s\n", symbol);
  fprintf(file, "\n");
  fprintf(file, "--\n");
  fprintf(file, "-- %s\n", filename);
  fprintf(file, "--\n");
  fprintf(file, "\n");
  fprintf(stderr, " finished\n");

  fprintf(stderr, "generating body...");
  fprintf(file, "#include \"../%s/port.smv\"\n", INCDIR);
  fprintf(file, "\n");
  fprintf(file, "MODULE node\n");
  fprintf(file, "\n");
  printVars(file, n, p, d);
  printTrans(file, n, p, d);
  fprintf(file, "\n");
  fprintf(file, "#endif\n");
  fprintf(stderr, " finished\n");

  fclose(file);

}


void printDifferent(FILE *file, int n, int p, int d) {

  int i, j;

  fprintf(file, "--------------------------------------------------------------------------------\n");
  fprintf(file, "-- different                                                                  --\n");
  fprintf(file, "--------------------------------------------------------------------------------\n\n");
  fprintf(file, "VAR\n  different: boolean;\n\n");
  fprintf(file, "TRANS\n  different -> next(different)\n\n");
  fprintf(file, "INVAR\n");
  for (i = 0; i < n - 1; i++) {
    for (j = i + 1; j < n; j++) {
      fprintf(file, "  (different -> ((node[%u].state = T3_root_1 & node[%u].state = T3_root_1) -> !(node[%u].rnd = node[%u].rnd))) %c\n",
              i, j, i, j, !((i == n - 2) & (j == n - 1)) ? '&' : ' ');
    }
  }
  fprintf(file, "\n");
  fprintf(file, "FAIRNESS\n  different\n\n");

} /* printDifferent */


void printMapInOut(FILE *file, int n, int p, int d) {

  int i, j;

  fprintf(file, "--------------------------------------------------------------------------------\n");
  fprintf(file, "-- map in out                                                                 --\n");
  fprintf(file, "--------------------------------------------------------------------------------\n\n");
  fprintf(file, "#define map_in_out(n, p) ( \\\n");
  fprintf(file, "  case \\\n");
  for (i = 0; i < n; i++) {
    fprintf(file, "    peer[n][p].node = %u: \\\n", i);
    fprintf(file, "      case \\\n");
    for (j = 0; j < p; j++) {
      fprintf(file, "        peer[n][p].port = %u: node[%u].port[%u].cout; \\\n", j, i, j);
    }
    fprintf(file, "        1: {idle, parent_notify, child_notify}; \\\n");
    fprintf(file, "      esac; \\\n");
  }
  fprintf(file, "    1: {idle, parent_notify, child_notify}; \\\n");
  fprintf(file, "  esac \\\n)\n\n");
  fprintf(file, "INVAR\n");
  for (i = 0; i < n; i++) {
    for (j = 0; j < p; j++) {
    fprintf(file, "  (!(peer[%u][%u].node = -1) -> node[%u].port[%u].cin in map_in_out(%u, %u)) %c\n", i, j, i, j, i, j, ((i == n - 1) && (j == p - 1)) ? ' ' : '&');
    }
  }
  fprintf(file, "\n");

} /* printMapInOut */


void printSoundConfig(FILE *file, int n, int p, int d) {

  int i, j;

  fprintf(file, "--------------------------------------------------------------------------------\n");
  fprintf(file, "-- sound configuration                                                        --\n");
  fprintf(file, "--------------------------------------------------------------------------------\n\n");
  fprintf(file, "#define sound_config(n, p) ( \\\n");
  fprintf(file, "  !(peer[n][p].node = n) & \\\n");
  fprintf(file, "  case \\\n");
  for (i = 0; i < n; i++) {
    fprintf(file, "    peer[n][p].node = %u: \\\n", i);
    fprintf(file, "      case \\\n");
    for (j = 0; j < p; j++) {
      fprintf(file, "        peer[n][p].port = %u: peer[%u][%u].node = n & peer[%u][%u].port = p; \\\n", j, i, j, i, j);
    }
    fprintf(file, "        1: 1; \\\n");
    fprintf(file, "      esac; \\\n");
  }
  fprintf(file, "    1: 1; \\\n");
  fprintf(file, "  esac \\\n)\n\n");
  fprintf(file, "INVAR\n");
  for (i = 0; i < n; i++) {
    for (j = 0; j < p; j++) {
      fprintf(file, "  sound_config(%u, %u) %c\n", i, j, (i == n - 1) && (j == p - 1) ? ' ' : '&');
    }
  }
  fprintf(file, "\n");

} /* printSoundConfig */


void printExistsPath(FILE *file, int n1, int n2, int l, int n, int p, int *visited) {

  int i, j;

  if (l < 0) {
    fprintf(file, "0");
  }
  else if (l == 0) {
    if (n1 == n2)
      fprintf(file, "1");
    else
      fprintf(file, "0");
  }
  else if (l == 1) {
    fprintf(file, "(");
    for (j = 0; j < p; j++) {
      fprintf(file, "peer[%u][%u].node = %u %c", n1, j, n2, (j == p - 1) ? ' ' : '|');
    }
    fprintf(file, ")");
  }
  else {
    fprintf(file, "(");
    for (i = 0; i < n; i++) {
      if (!visited[i]) {
	visited[i] = 1;
	fprintf(file, "(");
	printExistsPath(file, n1, i, 1, n, p, visited);
	fprintf(file, " & ");
	printExistsPath(file, i, n2, l - 1, n, p, visited);
	fprintf(file, ")");
	fprintf(file, " | ");
	visited[i] = 0;
      }
    }
    fprintf(file, "0");
    fprintf(file, ")");
  }
  
} /* printExistsPath */


void printNoCycle(FILE *file, int n, int p, int d) {

  int i, i2, i3, j, j2, l;
  int visited[MAXN];

  for (i = 0; i < n; i++)
    visited[i] = 0;
  fprintf(file, "--------------------------------------------------------------------------------\n");
  fprintf(file, "-- no cycle                                                                   --\n");
  fprintf(file, "--------------------------------------------------------------------------------\n\n");
  fprintf(file, "INVAR\n");
  fprintf(file, "  !(");
  for (i = 0; i < n; i++) {
    fprintf(file, "(");
    for (i2 = 0; i2 < n; i2++) {
      if (i2 != i) {
	for (j = 0; j < p; j++) {
	  fprintf(file, "peer[%u][%u].node = %u & (", i, j, i2);
	  for (i3 = 0; i3 < n; i3++) {
	    if (i3 != i) {
	      for (j2 = 0; j2 < p; j2++) {
		fprintf(file, "peer[%u][%u].node = %u & !(peer[%u][%u].port = %u) & (%u = %u | (", i3, j2, i, i3, j2, j, i2, i3);
		for (l = 1; l <= n - 2; l++) {
		  visited[i] = visited[i2] = visited[i3] = 1;
		  printExistsPath(file, i2, i3, l, n, p, visited);
		  visited[i] = visited[i2] = visited[i3] = 0;
		  fprintf(file, " | ");
		}
		fprintf(file, "0)) | ");
	      }
	    }
	  }
	  fprintf(file, "0) | ");
	}
      }
    }
    fprintf(file, "0) | ");
  }
  fprintf(file, "0)\n");
  fprintf(file, "\n");

}


void printIsConnected(FILE *file, int n, int p, int d) {

  int i, l;
  int visited[MAXN];

  for (i = 0; i < n; i++)
    visited[i] = 0;
  fprintf(file, "--------------------------------------------------------------------------------\n");
  fprintf(file, "-- is connected                                                               --\n");
  fprintf(file, "--------------------------------------------------------------------------------\n\n");
  fprintf(file, "INVAR\n");
  for (i = 0; i < n - 1; i++) {
    fprintf(file, "  (");
    for (l = 1; l < n; l++) {
      visited[i] = visited[i + 1] = 1;
      printExistsPath(file, i, i + 1, l, n, p, visited);
      visited[i] = visited[i + 1] = 0;
      if (l < n - 1) {
	fprintf(file, " | ");
      }
    }
    fprintf(file, ")");
    if (i < n - 2) {
      fprintf(file, " & \n");
    }
  }
  fprintf(file, "\n\n");

} /* printIsConnected */


void printInitConnected(FILE *file, int n, int p, int d) {

  int i, j;

  fprintf(file, "--------------------------------------------------------------------------------\n");
  fprintf(file, "-- init connected                                                             --\n");
  fprintf(file, "--------------------------------------------------------------------------------\n\n");
  fprintf(file, "#define init_connected(n) ( \\\n");
  for (j = 0; j < p; j++) {
    fprintf(file, "  (peer[n][%u].node = -1 <-> !node[n].port[%u].connected) %c \\\n", j, j, (j != p - 1) ? '&' : ' ');
  }
  fprintf(file, ") \n\n");
  fprintf(file, "INVAR\n");
  for (i = 0; i < n; i++) {
    fprintf(file, "  init_connected(%u) %c \n", i, (i < n - 1) ? '&' : ' ');
  }
  fprintf(file, "\n");

} /* printInitConnected */


void generateModel(int n, int p, int d) {

  FILE *file;
  char filename[256], pathname[256], symbol[256];
  int i;

  sprintf(filename, "model_%u_%u_%x.smv", n, p, d);
  sprintf(pathname, "%s/%s", OBJDIR, filename);
  strcpy(symbol, filename);

  file = fopen(pathname, "w");
  if (file == NULL) {
    fprintf(stderr, "cannot open file %s for writing\n", filename);
    exit(1);
  }
  else
    fprintf(stderr, "generating file %s\n", filename);

  fprintf(stderr, "generating header...");
  for (i = 0; i < strlen(symbol); i++) {
    if (islower(symbol[i]))
      symbol[i] = toupper(symbol[i]);
    else if (symbol[i] == '.')
      symbol[i] = '_';
  }
  fprintf(file, "#ifndef %s\n", symbol);
  fprintf(file, "#define %s\n", symbol);
  fprintf(file, "\n");
  fprintf(file, "--\n");
  fprintf(file, "-- %s\n", filename);
  fprintf(file, "--\n");
  fprintf(file, "\n");
  fprintf(stderr, " finished\n");

  fprintf(stderr, "generating body...");
  printDifferent(file, n, p, d);
  printMapInOut(file, n, p, d);
  if ((d & CFN) != 0) {
    printSoundConfig(file, n, p, d);
    printNoCycle(file, n, p, d);
    printIsConnected(file, n, p, d);
  }
  printInitConnected(file, n, p, d);
  fprintf(stderr, " finished\n");

  fprintf(file, "#endif\n");
  fclose(file);

} /* generateModel */


void printTimeout(FILE *file, int n, int p, int d) {

  int i;

  fprintf(file, "timeout := (\n");
  for (i = 0; i < n; i++) {
    fprintf(file, "  node[%u].timeout %c\n", i, i < n - 1 ? '|' : ' ');
  }
  fprintf(file, ");\n");
  fprintf(file, "\n");

} /* printTimeout */


void printKnownProblems(FILE *file, int n, int p, int d) {

  int i, i2, j;

  fprintf(file, "#define orphan(\\\n");
  for (i = 0; i < n; i++)
    fprintf(file, "  node%u%c\\\n", i, (i != n - 1) ? ',' : ' ');
  fprintf(file, "  )\\\n");
  fprintf(file, "  (node0.state = T2_parent & \\\n");
  fprintf(file, "  (\\\n");
  for (j = 0; j < p; j ++) {
    fprintf(file, "   (node0.port[%u].role = unknown & node0.port[%u].cout = parent_notify) %c \\\n", j, j, (j != p - 1) ? '|' : ' ' );
  }
  fprintf(file, "  ) & \\\n");
  for (i = 1; i < n; i++)
    fprintf(file, "  node%u.state = S0_start %c \\\n", i, (i != n - 1) ? '&' : ' ');
  fprintf(file, ")\n\n");
  fprintf(file, "known_problems := (\n");
  for (i = 0; i < n; i++) {
    fprintf(file, "  orphan(node[%u],\n", i);
    for (i2 = 0; i2 < n; i2++)
      if (i2 != i)
	fprintf(file, "    node[%u]%c\n", i2, !((i2 == n - 1) || ((i == n - 1) && (i2 == n - 2))) ? ',' : ' ');
    fprintf(file, "  ) %c \n", i < n - 1 ? '|' : ' ');
  }
  fprintf(file, ");\n");
  fprintf(file, "\n");

} /* printKnownProblems */


void printInStateS0Start(FILE *file, int n, int p, int d) {

  int i;

  fprintf(file, "in_state_S0_start := (\n");
  for (i = 0; i < n; i++) {
    fprintf(file, "  node[%u].state = S0_start %c \n", i, i < n - 1 ? '&' : ' ');
  }
  fprintf(file, ");\n");
  fprintf(file, "\n");

} /* printInStateS0Start */


void printAtLeastOneRoot(FILE *file, int n, int p, int d) {

  int i;

  fprintf(file, "at_least_one_root := (\n");
  for (i = 0; i < n; i++) {
    fprintf(file, "  node[%u].root %c \n", i, i < n - 1 ? '|' : ' ');
  }
  fprintf(file, ");\n");
  fprintf(file, "\n");

} /* printAtLeastOneRoot */


void printAtMostOneRoot(FILE *file, int n, int p, int d) {

  int i, i2;

  fprintf(file, "at_most_one_root := (\n");
  for (i = 0; i < n; i++) {
    fprintf(file, "  (node[%u].root -> ( \n", i);
    for (i2 = 0; i2 < n; i2++)
      if (i2 != i)
	fprintf(file, "    !node[%u].root %c \n", i2, (((i != n - 1) && (i2 < n - 1)) || (i2 < n - 2)) ? '&' : ' ');
    fprintf(file, "  )) %c \n", i < n - 1 ? '&' : ' ');
  }
  fprintf(file, ");\n");
  fprintf(file, "\n");

} /* printAtMostOneRoot */


void printRolesDetermined(FILE *file, int n, int p, int d) {

  int i, j;

  fprintf(file, "#define determined(node) (\\\n");
  for (j = 0; j < p; j++) {
    fprintf(file, "  (!(node.port[%u].role = off) -> !(node.port[%u].role = unknown)) %c \\\n", j, j, (j != p - 1) ? '&' : ' ');
  }
  fprintf(file, ")\n\n");
  fprintf(file, "roles_determined := (\n");
  for (i = 0; i < n; i++) {
    fprintf(file, "  determined(node[%u]) %c \n", i, i < n - 1 ? '&' : ' ');
  }
  fprintf(file, ");\n");
  fprintf(file, "\n");

} /* printRolesDetermined */


void printLinksIdle(FILE *file, int n, int p, int d) {

  int i, j;

  fprintf(file, "#define idle(node) (\\\n");
  for (j = 0; j < p; j++) {
    fprintf(file, "  node.port[%u].cout = idle %c \\\n", j, (j != p - 1) ? '&' : ' ');
  }
  fprintf(file, ")\n\n");
  fprintf(file, "links_idle := (\n");
  for (i = 0; i < n; i++) {
    fprintf(file, "  idle(node[%u]) %c \n", i, i < n - 1 ? '&' : ' ');
  }
  fprintf(file, ");\n");
  fprintf(file, "\n");

} /* printLinksIdle */


void generateSpec(int n, int p, int d) {

  FILE *file;
  char filename[256], pathname[256], symbol[256];
  int i;

  sprintf(filename, "spec_%u_%u_%x.smv", n, p, d);
  sprintf(pathname, "%s/%s", OBJDIR, filename);
  strcpy(symbol, filename);

  file = fopen(pathname, "w");
  if (file == NULL) {
    fprintf(stderr, "cannot open file %s for writing\n", filename);
    exit(1);
  }
  else
    fprintf(stderr, "generating file %s\n", filename);

  fprintf(stderr, "generating header...");
  for (i = 0; i < strlen(symbol); i++) {
    if (islower(symbol[i]))
      symbol[i] = toupper(symbol[i]);
    else if (symbol[i] == '.')
      symbol[i] = '_';
  }
  fprintf(file, "#ifndef %s\n", symbol);
  fprintf(file, "#define %s\n", symbol);
  fprintf(file, "\n");
  fprintf(file, "--\n");
  fprintf(file, "-- %s\n", filename);
  fprintf(file, "--\n");
  fprintf(file, "\n");
  fprintf(stderr, " finished\n");

  fprintf(stderr, "generating body...");
  fprintf(file, "DEFINE\n");
  fprintf(file, "\n");
  printTimeout(file, n, p, d);
  printKnownProblems(file, n, p, d);
  printInStateS0Start(file, n, p, d);
  printAtLeastOneRoot(file, n, p, d);
  printAtMostOneRoot(file, n, p, d);
  printRolesDetermined(file, n, p, d);
  printLinksIdle(file, n, p, d);
  fprintf(stderr, " finished\n");

  fprintf(file, "#endif\n");
  fclose(file);

}


void printNodes(char * filename, char * pathname, int n, int p, int d) {

  int i, j;
  FILE *file;
  char symbol[256], s[256];

  strcpy(symbol, filename);
  file = fopen(pathname, "r");
  if (file != NULL) {
    do {
      fprintf(stderr, "file %s exists, overwrite? (yes/no)\n", filename);
      fscanf(stdin, "%s", s);
    } while (strcmp(s, "yes") != 0 && strcmp(s, "no") != 0);
    fclose(file);
    if (strcmp(s, "no") == 0) {
      exit(1);
    }
  }
  file = fopen(pathname, "w");
  if (file == NULL) {
    fprintf(stderr, "cannot open file %s for writing\n", filename);
    exit(1);
  }
  else
    fprintf(stderr, "generating file %s\n", filename);

  fprintf(stderr, "generating header...");
  for (i = 0; i < strlen(symbol); i++) {
    if (islower(symbol[i]))
      symbol[i] = toupper(symbol[i]);
    else if (symbol[i] == '.')
      symbol[i] = '_';
  }
  fprintf(file, "#ifndef %s\n", symbol);
  fprintf(file, "#define %s\n", symbol);
  fprintf(file, "\n");
  fprintf(file, "--\n");
  fprintf(file, "-- %s\n", filename);
  fprintf(file, "--\n");
  fprintf(file, "\n");
  fprintf(stderr, " finished\n");

  fprintf(stderr, "generating body...");
  fprintf(file, "INVAR\n");
  for (i = 0; i < n; i++) {
    for (j = 0; j < p; j++) {
      fprintf(file, "  peer[%u][%u].node = -1 & peer[%u][%u].port = 0 %c \n", i, j, i, j, ((i < n - 1) || (j < p - 1)) ? '&' : ' ');
    }
  }
  fprintf(stderr, " finished\n");

  fprintf(file, "#endif\n");
  fclose(file);

} /* printNodes */


void generateNodes(int n, int p, int d) {

  FILE *file, *file2;
  char filename[256], pathname[256], symbol[256];
  char filename2[256], pathname2[256];
  int i;

  sprintf(filename, "nodes_%u_%u_%x.smv", n, p, d);
  sprintf(pathname, "%s/%s", OBJDIR, filename);
  strcpy(symbol, filename);
  memset(filename2, 0, 256*sizeof(char));
  memset(pathname2, 0, 256*sizeof(char));

  file = fopen(pathname, "w");
  if (file == NULL) {
    fprintf(stderr, "cannot open file %s for writing\n", filename);
    exit(1);
  }
  else
    fprintf(stderr, "generating file %s\n", filename);

  fprintf(stderr, "generating header...");
  fprintf(file, "#include \"peer_%u_%u_%x.smv\"\n", n, p, d);
  fprintf(file, "#include \"node_%u_%u_%x.smv\"\n", n, p, d);
  fprintf(file, "\n");
  fprintf(file, "--\n");
  fprintf(file, "-- %s\n", filename);
  fprintf(file, "--\n");
  fprintf(file, "\n");
  fprintf(stderr, " finished\n");

  fprintf(stderr, "generating body...");
  fprintf(file, "MODULE main\n");
  fprintf(file, "\n");
  fprintf(file, "VAR\n");
  fprintf(file, "  peer: array 0..%u of array 0..%u of peer;\n", n - 1, p - 1);
  fprintf(file, "  node: array 0..%u of node;\n", n - 1);
  fprintf(file, "\n");
  fprintf(file, "#include \"model_%u_%u_%x.smv\"\n", n, p, d);
  fprintf(file, "\n");
  if (~d & CFN) {
    sprintf(filename2, "config_%u_%u_%x.smv", n, p, d);
    sprintf(pathname2, "%s/%s", INCDIR, filename2);
    file2 = fopen(pathname2, "r");
    if (file2 == NULL) {
      fprintf(stderr, "cannot open file %s for reading, generating template\n", filename2);
      fprintf(stderr, "fill in template, then rerun\n");
      printNodes(filename2, pathname2, n, p, d);
      exit(1);
    }
    fclose(file2);
    fprintf(file, "#define USECONFIG\n");
  }
  else {
    fprintf(file, "#undef USECONFIG\n");
  }
  if (~d & FRN) {
    fprintf(file, "INVAR\n");
    for (i = 0; i < n; i++) {
      fprintf(file, "  node[%u].force_root = 0 %c \n", i, i < n - 1 ? '&' : ' ');
    }
  }
  fprintf(file, "#ifdef USECONFIG\n");
  fprintf(file, "#include \"../%s\"\n", pathname2);
  fprintf(file, "#endif\n");
  fprintf(file, "\n");
  fprintf(file, "#include \"spec_%u_%u_%x.smv\"\n", n, p, d);
  fprintf(file, "#include \"../%s/general_spec.smv\"\n", INCDIR);
  fprintf(file, "\n");
  fprintf(file, "SPEC\n");
  fprintf(file, "  1 -- fill in specification here\n");
  fprintf(file, "\n");
  fprintf(stderr, " finished\n");

  fclose(file);

}



int main(int argc, char *argv[]) {

  int n = 0;
  int p = 0;
  int d = 0;
  int i = 1;

  while (i < argc) {

    /* number of nodes */
    if (strcmp(argv[i], "-n") == 0) {
      i++;
      if (sscanf(argv[i], "%i", &n) != 1) {
	displayUsage(argv[0]);
	exit(1);
      }
    }

    /* number of ports */
    else if (strcmp(argv[i], "-p") == 0) {
      i++;
      if (sscanf(argv[i], "%i", &p) != 1) {
	displayUsage(argv[0]);
	exit(1);
      }
    }

    /* force root non-determinism */
    else if (strcmp(argv[i], "-frn") == 0) {
      d |= FRN;
    }

    /* configuration non-determinism */
    else if (strcmp(argv[i], "-cfn") == 0) {
      d |= CFN;
    }

    else {
      displayUsage(argv[0]);
      exit(1);
    }

    i++;
  }

  generatePeer(n, p, d);
  generateNode(n, p, d);
  generateModel(n, p, d);
  generateSpec(n, p, d);
  generateNodes(n, p, d);

  exit(0);

}
